var canvas = document.getElementById("invaders");
var processing = new Processing(canvas, function(processing) {
  processing.size(400, 500);
  processing.background(0xFFF);

  var mouseIsPressed = false;
  processing.mousePressed = function () { mouseIsPressed = true; };
  processing.mouseReleased = function () { mouseIsPressed = false; };

  var keyIsPressed = false;
  processing.keyPressed = function () { keyIsPressed = true; };
  processing.keyReleased = function () { keyIsPressed = false; };

  with (processing) {

	/*
	 * 2017-06-23
	 * Developed by Patrick Heney
	 *
	 * This game will adapt to fill the canvas, regardless of size.
	 * for a larger more "classic arcade" feel, append ?width=400&height=600
	 * to the URL above.
	*/

	//  Known issues:
	//  Sometimes when invaders leave the screen (aborted invasion?)
	//  Sometimes invaders advance too much (overzealous?)
	//  At high levels (~12+) bottom rows bounce off edge 1 frame early
	//  Player not properly restricted from moving near the left side of screen
	//  Player not properly restricted from top half of screen

	var Controls = {
		MOVE_LEFT : 37,  //  left-arrow  
		MOVE_RIGHT : 39, //  right-arrow 
		MOVE_UP : 38,    //  up-arrow    
		MOVE_DOWN : 40,  //  down-arrow  
		FIRE : 32        //  space bar   
	};


	////    Game Parameters ////


	//  Player Data
	var playerSize = 30;
	var playerShotSize = 6;
	var playerSpeed = 250;
	var playerShotSpeed = 200;
	var startHealth = 3;
	var PlayerArea = {
		top: height / 2,
		bottom: height - playerSize,
		left: playerSize,
		right: width - playerSize
	};


	//  Enemy Data
	var enemySize  = 40;
	var enemyShotSize = 10;
	var enemyShotSpeed = 100;
	var enemyShotRate = 0.2; //  shots per second
	var enemyShotDamage = 1;
	var enemySpeed = 20; // steps per second
	var enemyShotVariance = 0.3;

	//  Game Data
	var score = null;
	var currentLevel = null;
	var targetFps = 30;
	var fixedOffset = null;
	var DegToRad = PI / 180;

	//  Debug Controls
	var showHitboxes = false;
	var teleportOnMouse = false;
	var destroyAll = true;


	////    Utilities       ////


	var Sign = function (value) {
		if (value === 0) { value = 1; }
		return value / abs(value);
	};

	var Clamp = function (value, mn, mx) {
		return max(0, min(mx, value));
	};

	var RandomColor = function () {
		var r = round (random(100,255));
		var g = round (random(100,255));
		var b = round (random(100,255));    
		return color (r, g, b);
	};

	var VaryColor = function (baseColor) {
		var rnd = random(255);
    
		var r = rnd * red(baseColor);
		var g = rnd * green(baseColor);
		var b = rnd * blue(baseColor);
    
		var mx = max(r,max(g,b));
		var ratio = rnd / mx;
    
		return color(r * ratio, g * ratio, b * ratio);
	};

	var RandomVelocity = function () {
		var direction = random(2 * PI);
		return new PVector (cos(direction), sin(direction));
	};

	var startMessage = "INVADERS";
	var endMessage = "GAME OVER";
	var levelMessage = "Level X!";
	var inputMessage = "Push S to Start";
	var waitMessage = null;
	var messageSize = 55;
	var boldFont = createFont("fantasy");
	var WaitMessage = function (message, submessage, action) {
		this.message = message;
		this.submessage = submessage;
		this.actionOnPlayerInput = null;
    
		if (typeof action !== null && typeof action === 'function') {
			this.actionOnPlayerInput = function () {
				action();
				waitMessage = null;
			};
		}
	};
	WaitMessage.prototype.Draw = function () {
		fill(color(255,255,0));
		textFont(boldFont);
		textSize(messageSize);
		textAlign(CENTER, CENTER);
		text(this.message,width/2,height/3);
		textSize(messageSize * 0.35);
		if (score !== null && score > 0) {
			text("" + score,width/2,height/3 + messageSize);
		}
		text(this.submessage,width/2,height/3 +2 * messageSize);
	};

	var paused = null;
	var pauseMessage = "Paused";
	var pauseResumeMessage = "Push P to Resume";
	var resumeMessage = null;


	////    Time    ////

	var timeTextSize = 10;
	var timeColor = color(255, 255, 0);
	var normalFont = createFont("sans-serif");

	var Time = function () {
		this.startTime = 0;
		this.endTime = 0;
		this.deltaTime = 0;
		this.fps = 0;
	};
	Time.prototype.Update = function () {
		this.endTime = millis();
		this.deltaTime = (this.endTime - this.startTime)/1000;
		this.fps = this.deltaTime > 0 ? 1 / this.deltaTime : "--";
		this.startTime = this.endTime;
	};
	Time.prototype.DrawDelta = function () {
		textFont(normalFont);
		fill(timeColor);
		textSize(timeTextSize);
		textAlign(RIGHT, CENTER);
		text(this.deltaTime*1000,5*timeTextSize, height - 2*timeTextSize);
		textAlign(LEFT, CENTER);
		text("delta (millis)",5.5*timeTextSize, height - 2*timeTextSize);
	};
	Time.prototype.DrawFps = function () {
		fill(timeColor);
		textSize(timeTextSize);
		textAlign(RIGHT, CENTER);
		text(this.fps, width - 5*timeTextSize, height - 2*timeTextSize);
		textAlign(LEFT, CENTER);
		text("fps", width - 4.5*timeTextSize, height - 2*timeTextSize);
	};


	////    GameObject  ////


	var GameObject = function (x,y,s) {
		this.SetPosition(x,y);
		this.SetSize(s);
		this.Reset();
		this.offset = random(1000);
	};
	GameObject.prototype.Reset = function () {
		this.SetVelocity(0,0);
		this.SetColor(0,255,255);
		this.SetFiring(false);
		this.SetHealth(3);
		this.flashExpire = -1;
	};
	GameObject.prototype.Update = function (deltaTime) {
		if (deltaTime <= 0) { return; }
		if (this.velocity === null || this.velocity === undefined) { return; }
		if (this.position === null || this.position === undefined) { return; }
		var deltaV = this.velocity.get();
		deltaV.mult(deltaTime);
		this.position.add(deltaV);
	};
	GameObject.prototype.Draw = function () {
		this.DrawUsing(this.color);
    
		if (this.flashExpire > millis()) {
			this.DrawUsing(this.flashColor);
		}
    
		if (showHitboxes) {
			this.DrawHitbox();
		}
	};
	GameObject.prototype.DrawUsing = function (c) {
		pushMatrix();
		translate(this.position.x, this.position.y);
		noStroke();
		fill(c);
		rect(-0.5*this.size,  -0.5 * this.size, this.size, this.size);
		popMatrix();
	};
	GameObject.prototype.DrawHitbox = function() {
		pushMatrix();
		translate(this.position.x, this.position.y);
    
		//  hitbox
		noFill();
		strokeWeight(1);
		stroke(255, 255, 0);
		rect(-0.5 * this.size, -0.5 * this.size, this.size, this.size);
		popMatrix();
	};
	GameObject.prototype.SetPosition = function (x,y) {
		if (this.position === null || this.position === undefined) { 
			this.position = new PVector(0,0); 
		}
		this.position.x = x;
		this.position.y = y;
	};
	GameObject.prototype.MatchPositionOf = function (gameObject) {
		this.SetPosition (gameObject.position.x, gameObject.position.y);
	};
	GameObject.prototype.SetVelocity = function (x,y) {
		if (this.velocity === null || this.velocity === undefined) { 
			this.velocity = new PVector(0,0); 
		}
		this.velocity.x = x;
		this.velocity.y = y;
	};
	GameObject.prototype.SetSize = function (s) {
		this.size = s;
	};
	GameObject.prototype.SetColor = function (r,g,b) {
		this.color = color(r,g,b);
	};
	GameObject.prototype.SetHealth = function (h) {
		this.health = h;
	};
	GameObject.prototype.Intersects = function (otherGameObject) {
		var rightOk = (this.position.x + 0.5 * this.size) < (otherGameObject.position.x - 0.5 * otherGameObject.size);
		var leftOk = (this.position.x - 0.5 * this.size) > (otherGameObject.position.x + 0.5 * otherGameObject.size);
		var topOk = (this.position.y + 0.5 * this.size) < (otherGameObject.position.y - 0.5 * otherGameObject.size);
		var bottomOk = (this.position.y - 0.5 * this.size) > (otherGameObject.position.y + 0.5 * otherGameObject.size);
    
		return !(rightOk || leftOk || topOk || bottomOk);
	};
	GameObject.prototype.IsOffScreen = function () {
		return this.position.x <= (0 + this.size) ||
			this.position.x >= (width - this.size) ||
			this.position.y <= (0 + this.size) ||
			this.position.y >= (height - this.size);
	};
	GameObject.prototype.SetFiring = function (firing) {
		this.isFiring = firing;    
	};
	GameObject.prototype.IsAlive = function () {
		return this.health > 0;
	};
	GameObject.prototype.Damage = function (d) {
		this.health -= d;
	};
	GameObject.prototype.Destroy = function () {
		this.health = 0;
	};
	GameObject.prototype.Flash = function (c, duration) {
		this.flashColor = c;
		this.flashExpire = millis() + 1000 * duration;
	};


	////    Recycler    ////


	var Recycler = function () {
		this.alive = null;
		this.dead = null;
		this.NewItem = null;
		this.Reset();
		this.SetColor(color(255, 0, 255));
	};
	Recycler.prototype.Reset = function () {
		this.alive = [];
		this.dead = [];
		this.spawned = [];
	};
	Recycler.prototype.Update = function (deltaTime) {
		for (var i = this.alive.length ; i > 0 ; i--) {
			var item = this.alive[i-1];
			item.SetFiring(false);
			if (!item.IsAlive()) {
				this.dead.push(item);
				this.alive.splice(i-1, 1);
				continue;
			}
			item.Update(deltaTime);
		}
	};
	Recycler.prototype.Draw = function () {
		for (var i = 0 ; i < this.alive.length ; i++) {
			this.alive[i].Draw();   
		}
	};
	Recycler.prototype.Get = function () {
		var item = null;
		if (this.dead.length === 0) {
			item = this.NewItem();
		} else {
			item = this.dead.pop();
		}
		item.SetColor(this.color);
    
		this.alive.push(item);
		item.Reset();
		item.SetColor(this.color);
		return item;
	};
	Recycler.prototype.IsEmpty = function () {
		return this.alive.length === 0;
	};
	Recycler.prototype.SetColor = function (c) {
		this.color = c;
	};
	Recycler.prototype.GetItems = function () {
		return this.alive;
	};
	Recycler.prototype.GetRandom = function () {
		var item = null;
		if (this.alive.length > 0) {
			item = this.alive[floor(random(0,this.alive.length))];
		}
		return item;
	};

	////    Particles   ////


	var particleSystem = null;
	var particleSpeed = 750;
	var stylizedParticles = false;

	var Particle = function () {
		GameObject.apply(this);
		this.damping = 0.8;
		this.expireSpeed = null;
	};
	Particle.prototype = Object.create(GameObject.prototype);
	Particle.prototype.constructor = Particle;
	Particle.prototype.Update = function (deltaTime) {
		GameObject.prototype.Update.call(this, deltaTime);
		this.velocity.mult(this.damping);
		if (this.velocity.mag() < this.expireSpeed) {
			this.SetHealth(0);
		}
	};
	Particle.prototype.SetVelocity = function (x,y) {
		GameObject.prototype.SetVelocity.call(this,x,y);
		if (stylizedParticles) {
			this.damping = 0.8;
		} else {
			this.damping = random(0.7, 0.9);
		}
		this.expireSpeed = random(0,0.5) * this.velocity.mag();
	};

	var CreateParticle = function () {
		var particle = new Particle(0,0,0);
    
		particle.SetSize(2);
		particle.SetColor(RandomColor());
    
		return particle;
	};
	var ExplodeAt = function (position, quantity, baseColor) {
		if (particleSystem === null || particleSystem === undefined) { return; }
    
		var item = null;
		var velocity = null;
		for (var i = 0 ; i < quantity ; i++) {
			item = particleSystem.Get();
			item.SetPosition(position.x, position.y);
			velocity = RandomVelocity();
			velocity.mult(particleSpeed * random());
			item.SetVelocity(velocity.x, velocity.y);
			item.SetSize(random(1,5));
			if (baseColor && random() < 0.9) {
				item.SetColor(VaryColor(baseColor));
			} else {
				item.SetColor(RandomColor());
			}
		}
	};


	////    Shots   ////


	var Shot = function () {
		GameObject.apply(this);
		this.mode = 0;
	};
	Shot.prototype = Object.create(GameObject.prototype);
	Shot.prototype.constructor = Shot;
	Shot.prototype.Update = function(deltaTime) {
		GameObject.prototype.Update.call(this, deltaTime);
		if (this.IsOffScreen()) {
			this.SetHealth(0);
		}
	};
	Shot.prototype.DrawUsing = function (c) {
		switch (this.mode) {
			case 0: //  standard
				this.DrawStandard(c); break;
			case 1: //  surge
				this.DrawSurge(c); break;
			case 2: //  spin
				this.DrawSpin(c); break;
			case 3: //  double
				this.DrawDouble(c); break;
			case 4: //  wiggle
				this.DrawWiggle(c); break;
		}
	};
	Shot.prototype.DrawStandard = function (c) {
		pushMatrix();
		translate(this.position.x, this.position.y);
		noStroke();
		fill(c);
		rect(-0.5*this.size,  -2 * this.size, this.size, 4 * this.size);
		popMatrix();
	};
	Shot.prototype.DrawSpin = function (c) {
		pushMatrix();
		translate(this.position.x, this.position.y);
		rotate(this.offset + (this.velocity.y/56) * DegToRad * millis()/this.size);
		noStroke();
		fill(c);
		rect(-0.5*this.size,  -2 * this.size, this.size, 4 * this.size);
		popMatrix();
	};
	Shot.prototype.DrawWiggle = function (c) {
		pushMatrix();
		translate(this.position.x + sin(this.offset + DegToRad * millis()) * this.size, this.position.y);
		noStroke();
		fill(c);
		rect(-0.5*this.size,  -2 * this.size, this.size, 4 * this.size);
		popMatrix();
	};
	Shot.prototype.DrawSurge = function (c) {
		pushMatrix();
		translate(this.position.x, this.position.y + sin(this.offset + DegToRad * millis()) * this.size);
		noStroke();
		fill(c);
		rect(-0.5*this.size,  -2 * this.size, this.size, 4 * this.size);
		popMatrix();
	};
	Shot.prototype.DrawDouble = function (c) {
		pushMatrix();
		translate(this.position.x, this.position.y);
		noStroke();
		fill(c);
		rect(-1.1*this.size,  -2 * this.size, this.size, 4 * this.size);
		rect(0.1*this.size,  -2 * this.size, this.size, 4 * this.size);
		popMatrix();
	};
	Shot.prototype.SetMode = function (m) {
		this.mode = m;
	};

	var CreateShot = function () {
		var shot = new Shot(0,0,0);
		shot.SetSize(playerShotSize);
		return shot;
	};

	var nextShotTime = null;
	var ScheduleNextShotTime = function (level) {
		var variance = 1 + enemyShotVariance * random (-1,1);
		var adjustedRate = enemyShotRate * level;
		var delay = 1000 / adjustedRate;
		return millis() + variance * delay;
	};


	////    Player      ////


	var Player = function (x,y,s) {
		GameObject.apply(this,arguments);
		this.nextShotTime = 0;
		this.input = {
			moveLeft : false,
			moveRight : false,
			moveUp : false,
			moveDown : false,
			fire : false
		};
		this.playerShots = null;
		this.mouseControl = false;
		this.mouseInput = null;
	};
	Player.prototype = Object.create(GameObject.prototype);
	Player.prototype.constructor = Player;
	Player.prototype.GetDrawColor = function (requestColor) {
    
		//  request color
		if (this.health === startHealth) { return requestColor; }
    
		//  orange
		if (this.health >= startHealth /2) { return color(255, 170, 0); }
    
		//  red
		if (this.health >=1) { return color(255, 0, 0); }
    
		//  invisible
		if (this.health <= 0) { return color(0,0,0,255); }
	};
	Player.prototype.DrawUsing = function (c) {
		if (this.health <= 0) { return; }
    
		noFill();
		stroke(this.GetDrawColor(c));
		strokeWeight(5);
		this.DrawShip();
	};
	Player.prototype.DrawShip = function () {
		pushMatrix();
		translate(this.position.x, this.position.y);
			beginShape();
			vertex(0,- 0.25 *this.size);
			vertex(-0.1 * this.size, -0.5 * this.size);
			vertex(-0.5 * this.size, 0.5 * this.size);
			vertex(0,0.25 *this.size);
			vertex(0.5 * this.size, 0.5 * this.size);
			vertex(0.1 * this.size, -0.5 * this.size);
			vertex(0,-0.25 * this.size);
			endShape();
		popMatrix();
	};
	Player.prototype.SetFiring = function (f) {
		GameObject.prototype.SetFiring.call(this, f);
		if (!f) {
			this.nextShotTime = millis() + 200;
		}
	};
	Player.prototype.ReadyToFire = function () {
		return millis() > this.nextShotTime;
	};
	Player.prototype.Input = function (code, down) {
		var input = this.input;
    
		//  movement
		if (code === Controls.MOVE_LEFT) {
			input.moveLeft = down;
		}
		if (code === Controls.MOVE_RIGHT) {
			input.moveRight = down;
		}
		if (code === Controls.MOVE_UP) {
			input.moveUp = down;
		}
		if (code === Controls.MOVE_DOWN) {
			input.moveDown = down;
		}
    
		//  weapon fire
		if (code === Controls.FIRE) {
			input.fire = down;
		}
	};
	Player.prototype.UpdateVelocity = function () {

		var v = new PVector(0,0);
    
		if (!this.mouseControl) {
			//  keyboard input
        
			if (this.input.moveLeft) {
				v.add(new PVector(-1,0));
			}
        
			if (this.input.moveUp) {
				v.add(new PVector(0,-1));
			}
        
			if (this.input.moveRight) {
				v.add(new PVector(1,0));
			}
        
			if (this.input.moveDown) {
				v.add(new PVector(0,1));
			}
		} else if (this.mouseInput !== null) {
			//  mouse input
			v = this.mouseInput.get();
			v.sub(this.position);
		}
		//  set player velocity based on input
    
		if (v.mag() > 0) {
			v.normalize();
			v.mult(playerSpeed);
			this.SetVelocity(v.x, v.y);
		}
	};
	Player.prototype.Update = function (deltaTime) {
		this.UpdateVelocity();
    
		if (this.input.fire &&
			this.ReadyToFire() &&
			this.playerShots.alive.length < max(2,ceil(currentLevel/7))) {
			this.SetFiring(true);
		}
    
		GameObject.prototype.Update.call(this, deltaTime);
		if (this.IsOffScreen()) {
			this.position.x = Clamp(this.position.x, PlayerArea.left, PlayerArea.right);
			this.position.y = Clamp(this.position.y, PlayerArea.top, PlayerArea.bottom);
		}
		this.velocity = new PVector(0,0);
	};
	Player.prototype.IsOffScreen = function () {
		return this.position.x <= PlayerArea.left ||
			this.position.x >= PlayerArea.right ||
			this.position.y <= PlayerArea.top ||
			this.position.y >= PlayerArea.bottom;
	};
	Player.prototype.CancelInput = function () {
		this.Input(Controls.MOVE_LEFT, false);
		this.Input(Controls.MOVE_RIGHT, false);
		this.Input(Controls.MOVE_UP, false);
		this.Input(Controls.MOVE_DOWN, false);
		this.Input(Controls.FIRE, false);
	};

	var player = null;
	var playerShots = null;
	var InitPlayer = function () {
		playerShots = new Recycler();
		playerShots.NewItem = CreateShot;
		playerShots.Reset();
		playerShots.SetColor(color(255, 255, 255));
    
		player = new Player(width/2, height - 1.5 * playerSize, playerSize);
		player.Reset();
		player.SetHealth(startHealth);
		player.playerShots = playerShots;
	};


	////    EnemyCraft  ////


	var EnemyCraft = function(x,y,s) {
		GameObject.apply(this,arguments);
		this.Flash(color(255, 255, 255), 0.1);
		this.nextMoveTime = 0;
		this.toggle = false;
	};
	EnemyCraft.prototype = Object.create(GameObject.prototype);
	EnemyCraft.prototype.constructor = EnemyCraft;
	EnemyCraft.prototype.DrawUsing = function (c) {
		pushMatrix();
		var alternate = currentLevel % 2 === 1;
    
		//  body
		noStroke();
		fill(c);
		translate(this.position.x, this.position.y);
		rect(-0.3 * this.size,  -0.3 * this.size, 0.6 * this.size, 0.6 * this.size);
    
		//  fangs
		if (this.toggle) {
			//  left open
			pushMatrix();
			translate(-0.5 * this.size, 0.5 * this.size - 0.25 * this.size);
			rect(0, 0, 0.3 * this.size, 0.25 * this.size);
			popMatrix();
        
			//  right open
			pushMatrix();
			translate(0.5 * this.size -0.3 * this.size, 0.5 * this.size - 0.25 * this.size);
			rect(0, 0, 0.3 * this.size, 0.25 * this.size);
			popMatrix();
		} else {
			//  left closed
			pushMatrix();
			translate(-0.4 * this.size, 0.5 * this.size - 0.15 * this.size);
			rect(0, 0, 0.3 * this.size, 0.25 * this.size);
			popMatrix();
        
			//  right closed
			pushMatrix();
			translate(0.1 * this.size, 0.5 * this.size - 0.15 * this.size);
			rect(0, 0, 0.3 * this.size, 0.25 * this.size);
			popMatrix();
		}
    
		//  left horns
		if (this.toggle) {
			//  up
			fill(c);
			pushMatrix();
			translate(-0.6 * this.size, -0.7 * this.size);
			rect(0, 0, 0.25 * this.size,  0.6 * this.size);
			popMatrix();
		} else {
			//  down
			fill(c);
			pushMatrix();
			translate(-0.6 * this.size, -0.4 * this.size);
			rect(0, 0, 0.25 * this.size,  0.4 * this.size);
			popMatrix();
		}
        
		//  right horns
		if (this.toggle ^ alternate) {
			//  up
			pushMatrix();
			translate(0.6 * this.size, -0.7 * this.size);
			rect(-0.25 * this.size,  0, 0.25 * this.size,  0.6 * this.size);
			popMatrix();
		} else {
			//  down
			pushMatrix();
			translate(0.6 * this.size, -0.4 * this.size);
			rect(-0.25 * this.size,  0, 0.25 * this.size,  0.4 * this.size);
			popMatrix();
		}
    
		//  robot eyes
		fill(255, 255, 0);
		rect(-0.4 * this.size, -0.15 * this.size, 0.75 * this.size, 0.25 * this.size);
		fill(0, 0, 0);
		if (this.toggle) {
			//  eyes left
			rect(-0.35 * this.size, -0.1* this.size, 0.2 * this.size, 0.2 * this.size);
			rect(0.05 * this.size, -0.1 * this.size, 0.2 * this.size, 0.2 * this.size);
		} else {
			//  eyes right
			rect(-0.25 * this.size, -0.1* this.size, 0.2 * this.size, 0.2 * this.size);
			rect(0.15 * this.size, -0.1 * this.size, 0.2 * this.size, 0.2 * this.size);
		}
    
		popMatrix();
	};
	EnemyCraft.prototype.Update = function (deltaTime) {
		if (millis() < this.nextMoveTime) { return; }
		this.nextMoveTime = millis() + this.moveDelay;
    
		GameObject.prototype.Update.call(this, deltaTime * this.size * 0.5);
		this.toggle = !this.toggle;
	};
	EnemyCraft.prototype.SetSize = function (s) {
		this.moveDelay = s * 50;
		this.nextMoveTime = millis() + 0.5 * this.moveDelay;
		GameObject.prototype.SetSize.call(this, s);
	};
	EnemyCraft.prototype.Damage = function (d) {
		GameObject.prototype.Damage.call(this, d);
		if (this.health <= 0) { 
			score += currentLevel; 
		}
	};
	EnemyCraft.prototype.DistanceOffScreen = function () {
		var left = this.position.x - 0.5 * this.size;
		var right = this.position.x + 0.5 * this.size;
		var top = this.position.y - 0.5 * this.size;
		var bottom = this.position.y + 0.5 * this.size;
    
		var distance = max(0, 0 - left);
		distance = max(distance, right - width);
		distance = max(distance, 0 - top);
		distance = max(distance, bottom - height);
		return distance;
	};
	var CreateEnemyCraft = function () {
		return new EnemyCraft(0,0,0);
	};

	var EnemyFleet = function () {
		Recycler.apply(this,arguments);
	};
	EnemyFleet.prototype = Object.create(Recycler.prototype);
	EnemyFleet.prototype.constructor = EnemyFleet;
	EnemyFleet.prototype.Update = function (deltaTime) {
    
		var leader = this.alive[0];
		if (!leader) { return; }
		var orignalMoveTime = leader.nextMoveTime;
    
		Recycler.prototype.Update.call(this, deltaTime);
    
		//  Do nothing further when the fleet does not move.
		var fleetMoved = leader.nextMoveTime > orignalMoveTime;
		if (!fleetMoved) { return; }
    
		var collidedWithWall = false;

		//  Have any of the creeps progressed far enough down
		//  the screen to end the game?
		var maxProgress = 0;
		var maxDistanceOffScreen = 0;
		for (var i = 0 ; i < this.alive.length ; i++) {
			collidedWithWall |= this.alive[i].IsOffScreen();
			var distanceOffScreen = this.alive[i].DistanceOffScreen();
			if (abs(distanceOffScreen) > abs(maxDistanceOffScreen)) {
				maxDistanceOffScreen = distanceOffScreen;
			}
			var progress = this.alive[i].position.y;
			if (progress > maxProgress) {
				maxProgress = progress;
			}
		}
    
		if (waitMessage === null && maxProgress > height * 0.8) {
			player.SetHealth(0);
		}
    
		//  When any of the creeps touch the wall, 
		//  all creeps change direction.
		if (collidedWithWall) {
			var v = leader.velocity;
			v.x *= -1;
			if (maxDistanceOffScreen !== 0) { 
				maxDistanceOffScreen += (1 * Sign(maxDistanceOffScreen));
			}
			var offscreenAdjustment = new PVector(maxDistanceOffScreen, 0);
        
			for (var i = 0 ; i < this.alive.length ; i++) {
				this.alive[i].SetVelocity(v.x, v.y);
				this.alive[i].position.add(offscreenAdjustment);
            
				//  All creeps advance toward the player
				//  whenever they hit the wall.
				if (waitMessage === null) {
					var position = this.alive[i].position;
					var size = this.alive[i].size;
					var y = position.y + max(20, 0.5 * size);
					this.alive[i].SetPosition(position.x, y);
				}
			}
		}
	};
	EnemyFleet.prototype.DestroyAll = function () {
		for (var i = 0 ; i < this.alive.length ; i++) {
			this.alive[i].SetHealth(0);   
		}
	};

	var enemyCraft = null;
	var enemyShots = null;
	var InitFleet = function () {
		player.SetHealth(max(player.health, startHealth));
    
		currentLevel++;
		enemyCraft = new EnemyFleet();
		enemyCraft.NewItem = CreateEnemyCraft;
		enemyCraft.Reset();
    
		enemyShots = new Recycler();
		enemyShots.NewItem = CreateShot;
		enemyShots.SetColor(color(0, 255, 0));
		enemyShots.Reset();
		nextShotTime = ScheduleNextShotTime(currentLevel);
    
		var craftSize = max(10, enemySize - 2 * currentLevel);
		var spacing = 2 * craftSize;
		var rows = max(1, floor(height * 0.35 / spacing));
		var cols = floor(0.75 * width / spacing);
		var health = max(1, (currentLevel * random ()) / 2);
		var shade = RandomColor();
		var craft = null;
		var offset = 0;
		var craftCount = cols;
		var initialDirection = random() > 0.5 ? 1 : -1;
		var fleetPosition = (width - (cols * spacing)) * random();
    
		//  setup the fleet
		for (var i = 0 ; i < rows ; i++) {
			if (i % 2 > 0) {
				//  offset columns on "odd" levels
				craftCount = cols -1;
				offset = fleetPosition + spacing * 0.5;
			} else {
				craftCount = cols;
				offset = fleetPosition;
			}
        
			//  Make sure the creeps are "on screen" otherwise they
			//  will immediately advance several rows..
			var extra = (craftSize +1) - (0.5 * spacing + offset);
			extra = max(0, extra);
        
			for (var j = 0 ; j < craftCount ; j++) {
				craft = enemyCraft.Get();
				craft.Reset();
				craft.SetSize(craftSize);
				craft.SetColor(shade);
				craft.SetHealth(health);
				craft.SetPosition((j + 0.5) * spacing + offset + extra, (i + 0.6) * spacing);
				craft.SetVelocity(initialDirection * (enemySpeed + currentLevel), 0);
			}
		}
	};


	////    Physics         ////


	var CheckCollissions = function (impactors, target) {
		if (!target.IsAlive()) { return; }
        
		for (var i = 0 ; i < impactors.length ; i++) {
			var impactor = impactors[i];
			if (!impactor.IsAlive()) { continue; }
			if (impactor.Intersects(target)) {
				target.Damage(impactor.health);
            
				if (target.IsAlive()) {
					target.Flash(color(255,255,255), 0.15);
				} else {
					target.SetFiring(false);
					ExplodeAt(target.position, 90, target.color);
				}
				impactor.Destroy();
				break;
			}
		}
	};

	var CrossCheckCollissions = function (impactors, targets) {
		for (var i = 0 ; i < targets.length ; i++){
			CheckCollissions (impactors, targets[i]);
		}
	};


	////    Main Program    ////

	var time = null;
	var InitTime = function () {
		time = new Time();
	};

	var InitGame = function () {
		particleSystem = new Recycler();
		particleSystem.NewItem = CreateParticle;
		currentLevel = 0;
		score = 0;
		player.SetHealth(startHealth);
		player.SetPosition(width/2, height - 2 *player.size);
		InitFleet();
	};

	var InitProgram = function () {
		InitTime();
		InitPlayer();
    
		//  put some invaders on the title screen
		InitGame();
		InitFleet();
    
		//  show title and wait for player to start game
		waitMessage = new WaitMessage (startMessage, inputMessage, InitGame);
	};

	var IsLevelOver = function () {
		//  are all enemy craft destroyed?
		//  are all projectiles destroyed?
		return enemyCraft.IsEmpty() && enemyShots.IsEmpty() && playerShots.IsEmpty();
	};

	var IsGameOver = function () {
		//  does the player have zero health remaining?
		return !player.IsAlive();
	};

	var DrawGame = function () {
		if (player !== null) { player.Draw(); }
		if (enemyCraft !== null) { enemyCraft.Draw(); }
		if (enemyShots !== null) { enemyShots.Draw(); }
		if (playerShots !== null) { playerShots.Draw(); }
		if (particleSystem !== null) { particleSystem.Draw(); }
	};

	var UpdateGame = function () {
		if (player !== null) {
			if (enemyCraft !== null && enemyCraft.IsEmpty()) {
				player.SetFiring(false);
			}
			if (player.isFiring) {
				player.SetFiring(false);
            
				var shot = playerShots.Get();
				shot.Reset();
				shot.MatchPositionOf(player);
				shot.SetVelocity(0, -(playerShotSpeed + 20*currentLevel));
            
				//  TODO
				//  Player shot type changes every 10 levels
				//  Player shot type skips the "random" mode (mode 0);
				shot.SetMode(floor((currentLevel-1) / 10) % 5);
			}
        
			player.Update(time.deltaTime);
		}
    
		if (enemyCraft !== null) { enemyCraft.Update(time.deltaTime); }
		if (enemyShots !== null) { enemyShots.Update(time.deltaTime); }
		if (playerShots !== null) { playerShots.Update(time.deltaTime); }
		if (particleSystem !== null) { particleSystem.Update(time.deltaTime); }
    
		if (enemyCraft !== null && 
			enemyShots !== null &&
			waitMessage === null && 
			millis() > nextShotTime) {
        
			nextShotTime = ScheduleNextShotTime(currentLevel);
			var firingCraft = enemyCraft.GetRandom();
        
			if (firingCraft !== null) {
				var homingRate = Clamp(currentLevel*0.10,0.3,0.9);
				var homing = (currentLevel > 3) && (random() < homingRate);
				var shot = enemyShots.Get();
				shot.SetSize(firingCraft.size * 0.2);
				shot.MatchPositionOf(firingCraft);
				if (homing) {
					var speed = enemyShotSpeed + 6 * currentLevel;
					var dir = player.position.get();
					dir.sub(firingCraft.position);
					dir.normalize();
					dir.x *= random(0.98,1.02) * currentLevel / (currentLevel + 5);
					dir.normalize();
					dir.mult(speed);
					shot.SetVelocity(dir.x, dir.y);
				} else {
					shot.SetVelocity(0, enemyShotSpeed + 6 * currentLevel);
				}
				shot.SetHealth(enemyShotDamage);
				var shotMode = (currentLevel-1) % 6;
				if (shotMode === 5) { shotMode = floor(random(5)); }
				shot.SetMode(shotMode);
				firingCraft.Flash(color(255, 0, 0), 0.1);
			}
		}
    
		if (playerShots !== null && enemyCraft !== null) {
			CrossCheckCollissions(playerShots.GetItems(), enemyCraft.GetItems());
		}
    
		if (enemyShots !== null && player !== null) {
			CheckCollissions(enemyShots.GetItems(), player);
		}
    
		if (enemyCraft !== null && player !== null) {
			CheckCollissions(enemyCraft.GetItems(), player);
		}
	};

	var DrawScore = function () {
		textFont(normalFont);
    
		fill(timeColor);
		var size = messageSize / 3;
		textSize(size);
    
		//  score
		textAlign(LEFT, CENTER);
		text(score, size, size);
    
		//  level
		textAlign(RIGHT, CENTER);
		text(currentLevel, width - size, size);
	};

	var showCC = false;
	var DrawCC = function () {
		textFont(normalFont);
		textSize(timeTextSize);
		fill(color(255,255,0));
		textAlign(CENTER,CENTER);
		text("Source code (c) 2017 by Patrick Heney", width/2, height - 2* timeTextSize);
	};

	angleMode = "radians";
	frameRate(targetFps);
	InitProgram();


	////    Processing  ////


	draw = function() {
		background(50,50,50);
		time.Update();
    
		if (!paused) { UpdateGame(); }
		DrawGame();
    
		if (resumeMessage !== null) {
			resumeMessage.Draw();
		}
    
		if (showCC) { DrawCC(); }
    
		if (waitMessage !== null) { 
			waitMessage.Draw();
			return;
		}
    
		if (IsGameOver()) {
			player.CancelInput();
			waitMessage = new WaitMessage (endMessage, inputMessage.replace("Start", "Continue"), InitGame, score);
			return;
		} else if (IsLevelOver()) {
			player.CancelInput();
			var message = levelMessage.replace("X", currentLevel + 1);
			waitMessage = new WaitMessage (message, inputMessage, InitFleet);
			return;
		}
    
		DrawScore();
    
		if (!showCC) {
			time.DrawDelta();
			time.DrawFps();
		}
	};

	keyPressed = function() {
    
		//  debug tool -- destroy all
		if (keyCode === 78 && destroyAll) {
			enemyCraft.DestroyAll();
		}
    
		//  mouse control
		if (keyCode === 77 && waitMessage === null) { 
			player.mouseControl = !player.mouseControl; 
			if (!player.mouseControl) { player.mouseInput = null; }
		}
    
		//  pause
		if (keyCode === 80 && waitMessage === null) { 
			if (!paused) {
				resumeMessage = new WaitMessage(pauseMessage, pauseResumeMessage, null);
			} else {
				resumeMessage = null;
			}
			paused = !paused;
		}
    
		//  show copyright
		if (keyCode === 67) { showCC = !showCC; }
    
		//  check for player response to 'wait' message
		if (waitMessage !== null) {
			if (keyCode === 83) {
				//  's' to start level / start game (keyCode 83)
				waitMessage.actionOnPlayerInput();
			}
			return;
		}
    
		//  check for player weapon fire
		//  check for player movement
		player.Input (keyCode, true);
	};

	keyReleased = function() {
		player.Input (keyCode, false);
	};

	mouseClicked = function() {
		if (teleportOnMouse) {
			//  mouseX, mouseY, mouseButton
			player.SetPosition(mouseX, mouseY);
		}
    
		//  emulate 's' to start level / start game (keyCode 83)
		//if (waitMessage !== null) { waitMessage.actionOnPlayerInput(); }
	};

	mouseMoved = function() {
		if (player.mouseControl) {
			player.mouseInput = new PVector(mouseX, mouseY);
		}
	};

  }
  if (typeof draw !== 'undefined') processing.draw = draw;
});